import random
import os
import numpy as np


def random_prices(n, min_price=1, max_price=10**9, strategy="random"):
    """
    Generate a sequence of prices using different strategies:
    - random: completely random prices
    - increasing: generally increasing trend with noise
    - decreasing: generally decreasing trend with noise
    - waves: price waves (goes up and down periodically)
    - steps: prices move in distinct steps
    - v_shape: prices decrease then increase (V shape)
    - inv_v_shape: prices increase then decrease (inverse V shape)
    """
    prices = []

    if strategy == "random":
        prices = [random.randint(min_price, max_price) for _ in range(n)]

    elif strategy == "increasing":
        base = np.linspace(min_price, max_price, n)
        noise = np.random.normal(0, (max_price - min_price) / 20, n)
        prices = [
            max(min_price, min(max_price, int(p + n)))
            for p, n in zip(base, noise)
        ]

    elif strategy == "decreasing":
        base = np.linspace(max_price, min_price, n)
        noise = np.random.normal(0, (max_price - min_price) / 20, n)
        prices = [
            max(min_price, min(max_price, int(p + n)))
            for p, n in zip(base, noise)
        ]

    elif strategy == "waves":
        periods = random.randint(3, max(3, n // 10))
        x = np.linspace(0, periods * 2 * np.pi, n)
        amplitude = (max_price - min_price) / 2
        middle = (max_price + min_price) / 2
        prices = [int(middle + amplitude * np.sin(t)) for t in x]

    elif strategy == "steps":
        num_steps = random.randint(4, max(4, n // 5))
        step_size = n // num_steps
        current = random.randint(min_price, max_price)
        for i in range(num_steps):
            step_length = step_size if i < num_steps - 1 else n - i * step_size
            if random.random() > 0.5:  # decide up or down
                next_price = random.randint(current, max_price)
            else:
                next_price = random.randint(min_price, current)
            prices.extend([current] * step_length)
            current = next_price

    elif strategy == "v_shape":
        mid = n // 2
        base1 = np.linspace(max_price, min_price, mid)
        base2 = np.linspace(min_price, max_price, n - mid)
        noise = np.random.normal(0, (max_price - min_price) / 20, n)
        noise1, noise2 = noise[:mid], noise[mid:]
        prices1 = [
            max(min_price, min(max_price, int(p + n)))
            for p, n in zip(base1, noise1)
        ]
        prices2 = [
            max(min_price, min(max_price, int(p + n)))
            for p, n in zip(base2, noise2)
        ]
        prices = prices1 + prices2

    elif strategy == "inv_v_shape":
        mid = n // 2
        base1 = np.linspace(min_price, max_price, mid)
        base2 = np.linspace(max_price, min_price, n - mid)
        noise = np.random.normal(0, (max_price - min_price) / 20, n)
        noise1, noise2 = noise[:mid], noise[mid:]
        prices1 = [
            max(min_price, min(max_price, int(p + n)))
            for p, n in zip(base1, noise1)
        ]
        prices2 = [
            max(min_price, min(max_price, int(p + n)))
            for p, n in zip(base2, noise2)
        ]
        prices = prices1 + prices2

    # Ensure we have exactly n prices
    return prices[:n]


def optimal_local_trades(prices):
    """
    Generate the optimal trades by finding all local minima and maxima pairs.
    This helps identify the ground truth for maximum profits with unlimited trades.
    Returns a list of (buy_idx, sell_idx) pairs.
    """
    trades = []
    i = 0
    while i < len(prices) - 1:
        # Find next local minimum
        while i < len(prices) - 1 and prices[i] >= prices[i + 1]:
            i += 1

        if i == len(prices) - 1:
            break

        buy_idx = i

        # Find next local maximum
        i += 1
        while i < len(prices) - 1 and prices[i] <= prices[i + 1]:
            i += 1

        sell_idx = i

        # Only add if profitable
        if prices[sell_idx] > prices[buy_idx]:
            trades.append((buy_idx, sell_idx))

        i += 1

    return trades


def enforced_optimal_trades(n, min_price=1, max_price=10**9, num_trades=10):
    """
    Generate prices where the optimal solution has exactly num_trades profitable trades.
    Each trade will have a significant profit to ensure they're part of the optimal solution.
    """
    prices = []
    trades_per_segment = n // (2 * num_trades)
    min_val = min_price

    for i in range(num_trades):
        # Generate decreasing segment
        start_val = random.randint(min_val + 10, max_price)
        end_val = random.randint(min_val, start_val - 10)
        segment = list(
            range(
                start_val,
                end_val,
                -max(1, (start_val - end_val) // trades_per_segment),
            )
        )[:trades_per_segment]
        prices.extend(segment)

        # Generate increasing segment
        start_val = end_val
        end_val = random.randint(start_val + 10, max_price)
        segment = list(
            range(
                start_val,
                end_val,
                max(1, (end_val - start_val) // trades_per_segment),
            )
        )[:trades_per_segment]
        prices.extend(segment)

    # Pad remaining elements if needed
    while len(prices) < n:
        prices.append(prices[-1])

    # If we have too many prices, truncate
    return prices[:n]


def adversarial_test(n, min_price=1, max_price=10**9):
    """
    Generate an adversarial test where each optimal K trade selection
    requires careful consideration.
    """
    # Create a sequence where each additional trade k+1 requires
    # reconsidering the trading strategy from k trades
    prices = []

    # Start with a small value
    current = min_price
    prices.append(current)

    # Create a pattern of local min-max pairs with increasing profits
    for i in range(1, n, 2):
        # Local minimum
        current = max(min_price, current - random.randint(5, 20))
        prices.append(current)

        # Local maximum with increasing profit for each pair
        profit = random.randint(10, 30) * (i // 2 + 1)
        current = min(max_price, current + profit)
        if i + 1 < n:
            prices.append(current)

    # If n is odd, add one more price
    if len(prices) < n:
        prices.append(max(min_price, prices[-1] - random.randint(5, 20)))

    return prices[:n]


def alternating_profits_test(n, min_price=100, max_price=10**9):
    """
    Generate a test where the optimal profit for K trades alternates
    between different strategies, making it challenging for greedy approaches.
    """
    # Create a sequence of trades where:
    # - Some short trades have small profits but are quick
    # - Some long trades have large profits but span many prices
    prices = []
    current = random.randint(min_price, min_price * 10)
    prices.append(current)

    i = 1
    while i < n:
        # Short trade (small profit)
        current = max(min_price, current - random.randint(10, 50))
        if i < n:
            prices.append(current)
            i += 1

        current = current + random.randint(20, 100)
        if i < n:
            prices.append(current)
            i += 1

        # Long trade setup (large profit)
        current = max(min_price, current - random.randint(100, 500))
        if i < n:
            prices.append(current)
            i += 1

        # Add some fluctuation in the middle
        for _ in range(min(5, n - i)):
            current = max(min_price, current + random.randint(-30, 50))
            if i < n:
                prices.append(current)
                i += 1

        # Complete the large profit trade
        current = current + random.randint(400, 1000)
        if i < n:
            prices.append(current)
            i += 1

    return prices[:n]


def zigzag_ascending_test(n, min_price=1, max_price=10**9):
    """
    Generate a test with pattern like: 1, 999, 3, 998, 5, 997, ...
    With random noise added to segments.
    """
    prices = []
    low_val = min_price
    high_val = max_price
    
    # Determine how many values to generate before adding noise
    base_length = min(n, 2 * (max_price - min_price))
    
    # Generate the zigzag pattern
    for i in range(base_length):
        if i % 2 == 0:
            # Low value (ascending)
            prices.append(low_val)
            low_val += 2  # Increment by 2 for next low value
        else:
            # High value (descending)
            prices.append(high_val)
            high_val -= 1  # Decrement by 1 for next high value
            
        # Break if we've reached the limits
        if low_val >= high_val:
            break
    
    # Add random noise to fill up to n if needed
    while len(prices) < n:
        # Add segments of random noise
        segment_length = min(random.randint(5, 30), n - len(prices))
        segment_min = min_price
        segment_max = max_price
        
        # Generate random values for the segment
        for _ in range(segment_length):
            prices.append(random.randint(segment_min, segment_max))
    
    return prices[:n]


def zigzag_descending_test(n, min_price=1, max_price=10**9):
    """
    Generate a test with pattern like: 1, 1000, 2, 999, 3, 998, ...
    With random noise added to segments.
    """
    prices = []
    low_val = min_price
    high_val = max_price
    
    # Determine how many values to generate before adding noise
    base_length = min(n, 2 * (max_price - min_price))
    
    # Generate the zigzag pattern
    for i in range(base_length):
        if i % 2 == 0:
            # Low value (ascending)
            prices.append(low_val)
            low_val += 1  # Increment by 1 for next low value
        else:
            # High value (descending)
            prices.append(high_val)
            high_val -= 1  # Decrement by 1 for next high value
            
        # Break if we've reached the limits
        if low_val >= high_val:
            break
    
    # Add random noise to fill up to n if needed
    while len(prices) < n:
        # Add segments of random noise
        segment_length = min(random.randint(5, 30), n - len(prices))
        segment_min = min_price
        segment_max = max_price
        
        # Generate random values for the segment
        for _ in range(segment_length):
            prices.append(random.randint(segment_min, segment_max))
    
    return prices[:n]


def write_test_case(filename, prices):
    """Write a test case to the given filename."""
    with open(filename, "w") as f:
        f.write(f"{len(prices)}\n")
        f.write(" ".join(map(str, prices)))


def generate_test_suite():
    """Generate a suite of test cases with various strategies."""
    # Configurations: (n, strategy, count)
    test_configs = [
        # Group 1: n <= 20
        (10, "random", 2),
        (15, "increasing", 1),
        (20, "decreasing", 1),
        (10, "waves", 1),
        (15, "steps", 1),
        (20, "v_shape", 1),
        (15, "inv_v_shape", 1),
        (10, "enforced", 2),
        (20, "adversarial", 2),
        (15, "alternating", 1),
        (20, "zigzag_ascending", 1),
        # Group 3: n <= 2000
        (1000, "random", 2),
        (1500, "increasing", 1),
        (2000, "decreasing", 1),
        (1200, "waves", 1),
        (1800, "steps", 1),
        (1000, "v_shape", 1),
        (1500, "inv_v_shape", 1),
        (2000, "enforced", 2),
        (1200, "adversarial", 2),
        (1800, "alternating", 1),
        (2000, "zigzag_descending", 1),
        # Group 4: n <= 10000
        (10000, "random", 2),
        (8000, "increasing", 1),
        (10000, "decreasing", 1),
        (3000, "waves", 2),
        (7000, "steps", 2),
        (5000, "v_shape", 2),
        (8000, "inv_v_shape", 2),
        (10000, "enforced", 2),
        (10000, "adversarial", 2),
        (10000, "alternating", 2),
        (10000, "zigzag_ascending", 1),
        (10000, "zigzag_descending", 1),
        (10000, "zigzag_ascending", 1),
        # Group 5: n <= 100000
        (50000, "random", 2),
        (75000, "increasing", 2),
        (100000, "decreasing", 2),
        (30000, "waves", 2),
        (60000, "steps", 2),
        (90000, "v_shape", 2),
        (40000, "inv_v_shape", 2),
        (70000, "enforced", 2),
        (85000, "adversarial", 2),
        (95000, "alternating", 2),
        (100000, "enforced", 1),
        (100000, "adversarial", 1),
        (100000, "alternating", 1),
        (100000, "enforced", 1),
        (100000, "random", 1),
        (100000, "waves", 1),
        (100000, "zigzag_ascending", 1),
        (100000, "zigzag_descending", 1),
        (100000, "zigzag_ascending", 1),
        # Group 6: n <= 500000
        (200000, "random", 2),
        (300000, "increasing", 2),
        (400000, "decreasing", 2),
        (500000, "waves", 2),
        (250000, "steps", 2),
        (350000, "v_shape", 2),
        (450000, "inv_v_shape", 2),
        (200000, "enforced", 2),
        (300000, "adversarial", 2),
        (400000, "alternating", 2),
        (500000, "random", 1),
        (500000, "increasing", 1),
        (500000, "decreasing", 1),
        (500000, "waves", 1),
        (500000, "zigzag_ascending", 1),
        (500000, "zigzag_descending", 1),
        (500000, "zigzag_ascending", 1),
    ]

    test_num = 1
    for n, strategy, count in test_configs:
        for _ in range(count):
            if strategy == "enforced":
                prices = enforced_optimal_trades(n, num_trades=int(n / 10))
            elif strategy == "adversarial":
                prices = adversarial_test(n)
            elif strategy == "alternating":
                prices = alternating_profits_test(n)
            elif strategy == "zigzag_ascending":
                prices = zigzag_ascending_test(n, max_price=min(10**9, n))
            elif strategy == "zigzag_descending":
                prices = zigzag_descending_test(n, max_price=min(10**9, n))
            else:
                prices = random_prices(n, strategy=strategy)

            filename = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                f"stock.{str(test_num).zfill(2)}.in",
            )
            write_test_case(filename, prices)
            print(f"Generated {filename} - {strategy} with n={n}")
            test_num += 1


if __name__ == "__main__":
    generate_test_suite()
